import collections


def merge_dict(dict1: dict, dict2: dict) -> dict:
    """
    An actually good deep merge
    https://stackoverflow.com/a/61708681
    """
    q = collections.deque([(dict1, dict2)])
    while len(q) > 0:
        d1, d2 = q.pop()
        for k, v in d2.items():
            if k in d1 and isinstance(d1[k], dict) and isinstance(v, dict):
                q.append((d1[k], v))
            else:
                d1[k] = v

    return dict1



'''
Copyright (C) 2022 Aarnav Bos

This program is free software: you can redistribute it and/or modify

it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
import argparse
import glob
import os
from importlib import resources as pkg_resources
from typing import Dict, List, Union

import jinja2
import yamale
from box import Box
from yaml import load as load_yaml

from .colors import print_error, print_success
from .schema import config_schema, test_schema
from .util import merge_dict

try:
    from yaml import CLoader as YamlLoader
except ImportError:  # pragma: no cover
    from yaml import Loader as YamlLoader


def get_template(path: str, is_builtin: bool) -> jinja2.Template:
    if is_builtin:
        return jinja2.Template(
            pkg_resources.read_text('parserkiosk.templates', f'{path}.jinja2')
        )
    with open(path, 'r') as template_file:
        return jinja2.Template(template_file.read())


def get_ext(template_name: str) -> str:
    match template_name:
        case 'node_js':
            # jest needs a .test prefix
            return 'test.js'
        case 'python':
            return 'py'
        case _:
            raise Exception(f'Unsupported builtin template "{template_name}".')


def generate_test(
    filename: str,
    tests: Box,
    config: Box,
    template: jinja2.Template,
    ext: str,
) -> None:
    try:
        with open(f'tests/{filename}.{ext}', 'w') as test_file:
            test_file.write(
                template.render(
                    tests=tests.tests,
                    func=(
                        config.serialize_function
                        if tests.type == 'SERIALIZE'
                        else config.de_serialize_function
                    ),
                    import_string=config.import_string,
                    assert_funcs=config.assert_functions,
                    is_dict=lambda x: isinstance(x, Box),
                )
            )
    except FileNotFoundError:
        os.mkdir(os.path.join(os.path.curdir, 'tests'))
        return generate_test(filename, tests, config, template, ext)


def read_yaml(filename, validation_schema=None) -> Box:
    with open(filename, 'r') as file:
        content = file.read()
        if validation_schema:
            data = yamale.make_data(content=content)
            yamale.validate(validation_schema, data)
            # returning yaml_dict from yamale internal rep.
            return data[0][0]
    return load_yaml(content, Loader=YamlLoader)


def validate_cli_args(args) -> None:
    if not os.path.isdir(args.dir):
        raise Exception('The dir provided is not a directory!')
    if not os.path.exists(os.path.join(args.dir, 'config.yaml')):
        raise Exception('No config found in directory provided')
    if args.builtin and args.path:
        raise Exception('Cannot generate two test-suites at a time.')
    elif not args.builtin and not args.path:
        raise Exception('Need a template to generate a test-suite.')
    if args.path and not args.ext:
        raise Exception(
            'Need an --ext parameter when using a custom template.'
        )


def parse_tests(
    tests: List[str],
    override: bool,
) -> Union[None, List[Dict[str, Union[str, Box]]]]:
    parsed_tests = []
    for test_file in tests:
        test_yaml = read_yaml(test_file)
        _is_overriden = False
        if override and os.path.exists(
            test_file.replace('.yaml', '.override.yaml')
        ):
            _is_overriden = True
            override_test_yaml = read_yaml(
                test_file.replace('.yaml', '.override.yaml')
            )
            merge_dict(test_yaml, override_test_yaml)
        try:
            yamale.validate(test_schema, [(test_yaml, test_file)])
            parsed_tests.append(
                {
                    'name': test_file.split('/')[-1].replace('.yaml', ''),
                    'tests': Box(test_yaml),
                }
            )
        except yamale.YamaleError as e:
            print_error(str(e))
            if _is_overriden:
                print_error(
                    'Note: The file is being overriden as you are using the --override option and a corresponding .override.yaml file exists'  # noqa E501
                )
            return
    return parsed_tests


def main() -> None:
    parser = argparse.ArgumentParser(
        description='''Generate test suites for parsers in multiple languages.'''  # noqa E501
    )
    parser.add_argument(
        'dir',
        action='store',
        help='Your directory containing the test declaration and configuration',  # noqa E501
    )
    parser.add_argument(
        '--path',
        action='store',
        required=False,
        help='Path to your custom jinja2 template.',
    )
    parser.add_argument(
        '--ext',
        action='store',
        required=False,
        help='The extension for your test file(s), only needed for custom templates',  # noqa E501
    )
    parser.add_argument(
        '--builtin',
        action='store',
        required=False,
        help='Use a built-in templates: python(Python/pytest), node_js(Node.JS/jest).',  # noqa E501
    )
    parser.add_argument(
        '--override',
        action='store_true',
        required=False,
        help='Allow overriding a test files',
    )
    args = parser.parse_args()
    try:
        validate_cli_args(args)
        ext = args.ext if args.ext else get_ext(args.builtin)
    except Exception as e:
        print_error(e)
        return
    test_files = glob.glob(f'{args.dir}/test_*.yaml')
    test_files = [file for file in test_files if '.override.yaml' not in file]
    if len(test_files) == 0:
        print('No test files found. in {args.dir}')
        return
    template = get_template(
        args.builtin or args.path, args.builtin is not None
    )
    try:
        config = Box(
            read_yaml(
                os.path.join(args.dir, 'config.yaml'),
                config_schema,
            )
        )
    except yamale.YamaleError as e:
        print_error('Error(s) in config.yaml:')
        print_error(str(e))
        return
    if parsed_tests := parse_tests(test_files, override=args.override):
        for test in parsed_tests:
            generate_test(
                test.get('name'),
                test.get('tests'),
                config,
                template,
                ext,
            )
        print_success('Done')



'''
Copyright (C) 2022 Aarnav Bos

This program is free software: you can redistribute it and/or modify

it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
'''
from functools import reduce
from importlib import resources as pkg_resources
from typing import Any, Dict

from yamale import make_schema
from yamale.validators import DefaultValidators, Validator


class TestAssertValidator(Validator):
    tag = 'assert_validator'
    _errors = []

    def _is_valid(self, value: Dict[str, Any]) -> bool:
        func = value.get('func')
        arg = value.get('arg')
        if func is None:
            self._errors.append('Missing "func"')
        if len(value.keys()) > 2:
            self._errors.append(
                'Too many keys provided. Need only "func" and "arg"'
            )
        if func != 'fail' and arg is None:
            self._errors.append(
                'Need to provide an "arg" if "func" is not "fail"'
            )
        else:
            if type(arg) == dict:
                if not arg.get('type'):
                    self._errors.append('Missing "type" in assert.arg')
                if arg.get('type') not in ['raw', 'file', 'str']:
                    self._errors.append(
                        'Invalid value for "type". Allowed: "raw", "file" or "str"'  # noqa E501
                    )
                if not arg.get('arg'):
                    self._errors.append('Missing "arg"')
        return len(self._errors) == 0

    def fail(self, value):
        error_text = reduce(lambda a, b: f'{a}\n{b}', self._errors)
        self._errors = []
        return error_text


config_schema = make_schema(
    content=pkg_resources.read_text(
        'parserkiosk.schemas', 'config_schema.yaml'
    )
)

validators = DefaultValidators.copy()
validators[TestAssertValidator.tag] = TestAssertValidator

test_schema = make_schema(
    content=pkg_resources.read_text('parserkiosk.schemas', 'test_schema.yaml'),
    validators=validators,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['parserkiosk', 'parserkiosk.schemas', 'parserkiosk.templates']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'python-box>=5.4.1,<6.0.0',
 'yamale>=4.0.4,<5.0.0']

entry_points = \
{'console_scripts': ['parserkiosk = parserkiosk:main']}

setup_kwargs = {
    'name': 'parserkiosk',
    'version': '0.3.0',
    'description': 'A Proof of Concept multi-lingual test generation suite intended for parsers.',
    'long_description': '## ParserKiosk: A Proof of Concept multi-lingual test generation suite intended for parsers\n[![Coverage Status](https://coveralls.io/repos/github/R9295/parserkiosk/badge.svg)](https://coveralls.io/github/R9295/parserkiosk)\n![](https://img.shields.io/github/commit-activity/w/R9295/parserkiosk?style=flat-square)\n![](https://img.shields.io/github/issues/R9295/parserkiosk?style=flat-square)\n![](https://img.shields.io/pypi/v/parserkiosk?style=flat-square)\n[![Downloads](https://pepy.tech/badge/parserkiosk/week)](https://pepy.tech/project/parserkiosk)\n![](https://img.shields.io/pypi/format/parserkiosk?style=flat-square)\n![](https://img.shields.io/badge/code%20style-black-000000.svg)\n\n### Motivation\nAfter reading this [article](https://seriot.ch/projects/parsing_json.html) and [this one](https://bishopfox.com/blog/json-interoperability-vulnerabilities), I am now paranoid and under the assumption that implementations of data serialization and deserialization have a lot of quirks that differ from language to language, and implementation to implementation.\n\nThis _could_ lead to serious security issues as applications, especially web applicatons _usually_ utilize multiple services, written in multiple languages that use the same format to communicate. \n\nReference implementations usually provide tests, but translating them from language to language is tiresome and tedious. I wanted to compose a library to generate **simple**, functional tests for multiple languages with minimal repitition. \n\n### Usage\n1. Install \n``` bash\n$ pip install parserkiosk\n```\n2. Define a ``config.yaml``\n``` yaml\n# config.yaml\n---\nimport_string: "from my_parser import serialize, deserialize"\nserialize_function: "serialize"\nde_serialize_function: "deserialize"\nassert_functions:\n  - my_assert_function\n```\n3. Define a yaml file prefixed with ``test_`` in the same directory as ``config.yaml``\n``` yaml\n# test_serialize.yaml\ntype: "SERIALIZE"\ntests:\n  test_something:\n      info: "Example Test"\n      input:\n        type: "str"\n        arg: "hello, world"\n      assert:\n        func: "my_assert_function"\n        arg: "[\\"hello\\", \\" world\\"]"\n```\n4. Run parserkiosk in the same directory as ```config.yaml``` and ``test_serialize.yaml``\n``` bash\n$ parserkiosk . --builtin python\n```\n5. See output directory ```tests/```\n``` bash\n$ ls tests/\ntest_serialize.py\n```\n\n\n#### See [HOWTO](HOWTO.md) for a complete guide.\n\n### How does it work?\nParserkiosk uses ``jinja2`` templates to generate test cases from ``yaml`` file(s). You can either expect something to fail(raise an "exception" or "error") or use a function that you define in a special file called ```commons``` to assert if the parsed data matches the expected internal representation. \n\nLet\'s say you\'ve written a reference implementation of your data de/serialization format in ``Go`` and I wanted to implement it in ``Python``.  \nAll I would need to do to implement your test-suite is:\n1. Write a ```commons.py``` file implementing the same assertion functions that you\'ve written in your ``commons.go`` file\n2. Adapt your parserkiosk config to use my implementation\n3. Run ```$ parserkiosk folder_with_yaml_test_files/ --builtin python``` and _voila_ I have your entire test suite implemented!\n\nFor more on this, see ```examples/json/```\n\n### Languages supported\n- [x] Python / pytest / ``python``\n- [x] NodeJS / jest (sync) / ``node_js``\n- [ ] NodeJS / jest (async)\n- [ ] Lua \n- [ ] Go\n- [ ] Java\n- [ ] PHP\n- [ ] Perl\n- [ ] Ruby\n- [ ] ...\n\n### License\nAll work is licensed under ```GPL-3.0``` excluding the example JSON test-suite which is licensed under ```MIT```\n\n### Contributing\nIssues, feedback and pull requests are welcome. I have tried _my best_ to keep the code simple. Please keep in mind that I wish to limit features that we accomodate to keep it simple. Tests should be simple and readable.\n\n### Installing for development\n``` bash\n$ git clone https://github.com/you/your-fork-of-parserkiosk.git\n$ cd your-fork-of-parserkiosk\n$ poetry install\n$ poetry run pre-commit install\n# do some changes\n$ ./runtests.sh\n# you are ready!\n```\n### Thanks\nSpecial thanks to [nst](https://github.com/nst/) for inspiring Parserkiosk. All test cases in ``examples/json`` come from his [incredible work](https://github.com/nst/JSONTestSuite)\n',
    'author': 'aarnav',
    'author_email': 'aarnavbos@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/R9295/parser-kiosk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
